/*
 * This program returns the buffer sizes (underflow/overflow)
 * required for the given stream. The stream data is encoded
 * as produced by the scanvbr code. To compute the overflow
 * buffer requirements, this module simulates the client 
 * buffer during transmission.
 *
 * Juan Miguel del Rosario, (c) 1995.
 * All rights reserved.
 */

static char rcsid[] = "$Id: stats.c,v 1.3 1995/07/13 04:14:41 mrosario Exp mrosario $";

/* $Log: stats.c,v $
 * Revision 1.3  1995/07/13  04:14:41  mrosario
 * Modified to run on SGI. All irint() related variables
 * changed to rint() (i.e., from int to double). Major mod.
 * Also, floor() no longer taken for framesPerInterval. -- jmvr
 *
 * Revision 1.2  1995/06/24  17:57:18  mrosario
 * Modified to accept an arbitrary transmission interval.
 * The buffer computation remains conservative.   jmvr
 *
 * Revision 1.1  1995/06/24  15:48:35  mrosario
 * Initial revision
 * */

 
#include <stdio.h>
#include <string.h>

#define FPS 30.0

/*---- Globals -----*/ 
  struct Queue{
    int fsize;
    int fno;
    struct Queue *next;
  };

  struct Queue *Current = NULL;
  struct Queue *Qhead = NULL;
  int ConsumedFrames = 0;
  double framesPerInterval;

/*-------------------*/

double computeOflow();
double getOverflow();

/*----------< MAIN >----------------*/

#ifdef NOTSELFRUN
stats(m, filename, fnameout, interval)
   int m;
   char *filename;
   char *fnameout;
   double interval;    /* transmission interval in msecs. */
#else
main(argc, argv) 
   int argc;
   char *argv[];
#endif
{
  int cnt;
  FILE *fd, *fdo;
  double maxOflow, maxUflow;
  int maxOflowSlots, maxUflowSlots;
  int fnum, fsize;
  double framecount;
  double delta;
  int OflowSlots, UflowSlots;
  int i, buffsize;
  double dd;
  double rate, time;
  double timeslot, rem, tmp;
  double sendInterval;

  maxOflow = maxUflow = 0;
  maxOflowSlots = maxUflowSlots = 0;
  OflowSlots = UflowSlots = 0;
  framecount = 0;
  i = 0;
  delta = 0;
  timeslot = 0;
  rem = 0;

#ifdef NOTSELFRUN
  fd = fopen(filename, "r");
  fdo = fopen(fnameout, "a+");
  sendInterval = interval/1000.0;        /* convert transmission interval to secs. */
#else
  if (argc < 3) {
    printf("Usage: stats [filename] [send interval (msecs)]\n");
    exit();
  }
  fd = fopen(argv[1], "r");
  sendInterval = atoi(argv[2])/1000.0;   /* convert to secs. */
#endif 

  if (fd == 0) {
     printf("Error in opening file\n");
     exit();
   }

#ifdef DEBUG
  printf("Opened file\n");
#endif

 /*
  * Compute the number of frames consumed by the client in the 
  * interval of time selected (sendInterval). Note that the client
  * consumes frames in discrete steps (1 every 30th of a sec), so
  * we take the floor of the product of the time interval and the
  * frame rate. <-- OBSOLETE
  */ 
  framesPerInterval = FPS*sendInterval;

  printf("Transmission interval       : %.3e (secs)\n", sendInterval);
  printf("Frames consumed per Interval: %.3e\n", framesPerInterval);

  while (!feof(fd)) {
    if (timeslot < sendInterval) {
       fscanf(fd,"%d %d %lf %lf\n", &fnum, &fsize, &rate, &time);
       timeslot += time;
       framecount += 1.0;
     }

    if (timeslot >= sendInterval) {     /* ONE TIME INTERVAL HAS ELAPSED */
       rem = timeslot - sendInterval;   /* we may send partial frames */ 
       if (rem > 0) {                   /* But we don't count incomplete */
          framecount -= 1.0;            /*   frames as having been sent */
       }
       delta += framecount - framesPerInterval; /* Overflow or Underflow slot? */

#ifdef DEBUG
  printf("-----------------\n");
  printf("timeslot %.4f\n", timeslot); 
  printf("num %d, size %d, rate %.4f, time %.4f\n", fnum,fsize,rate,time);
  printf("time slot passed: rem %f, delta %.3e, fcount %.3e\n", rem, 
       delta, framecount);
  printf("-----------------\n");
#endif

      if (delta < maxUflow) {          /* delta < 0 => underflow */
        maxUflow = delta;              /* store largest underflow */
      }
      if (delta > maxOflow) {          /* delta > 0 => overflow */
        maxOflow = delta;              /* store largest overflow */
      }
      if (framecount > framesPerInterval){
        OflowSlots++;
      }
      if (framecount < framesPerInterval){
        UflowSlots++;
      }
      cnt = UflowSlots - OflowSlots;
      if (cnt > maxUflowSlots) {
        maxUflowSlots = cnt;
      }
      cnt = OflowSlots - UflowSlots;
      if (cnt > maxOflowSlots) {
        maxOflowSlots = cnt;
      }

      timeslot = rem;
      if (rem > 0) framecount = 1.0;
      else framecount = 0;
    }
  }

#ifndef NOTSELFRUN
  printf("\n");
  printf("OVERFLOW/UNDERFLOW-SLOT DATA:\n");
  printf("Overflow Slots       : %d\n", maxOflowSlots);
  printf("Underflow Slots      : %d\n", maxUflowSlots);
  printf("Max. Underflow Frames: %.6e\n", maxUflow);
  printf("Max. Overflow Frames : %.6e\n", maxOflow);
  printf("\n");
#endif

  fseek(fd,0,0);
  buffsize = 0;
  dd = 0;
  while (dd > maxUflow) {         /* maxUflow < dd < 0 */ 
    fscanf(fd,"%d %d %lf %lf\n", &fnum, &fsize, &rate, &time);
    buffsize += fsize;
    dd -= 1.0;
  }

#ifdef NOTSELFRUN
                                  /* returns MAXIMUM buffer requirment */
  fprintf(fdo,"%d  %d  %.3lf\n", m, buffsize, computeOflow(fd, maxUflow,
          sendInterval)); 
  fclose(fdo);
#else 
  printf("Underflow Buffer requirement : %.3e (bits), %.3e (bytes)\n", 
          buffsize*1.0, buffsize/8.0); 
  tmp = computeOflow(fd, maxUflow, sendInterval);
  printf("Maximum Buffer requirement : %.3e (bits), %.3e (bytes)\n", 
          tmp, tmp/8.0);
  printf("\n\n");
#endif 

  fclose(fd);
}

/*-----------------------------------------*/

double computeOflow(fd, Uflow, sendInterval)
  FILE *fd;
  double Uflow;     /* max underflow frames */
  double sendInterval;
{  
  struct Queue *tmpPtr = NULL;
  struct Queue *Qptr = NULL;
  struct Queue *PartFrameptr = NULL;
  int fno, fsize;
  double Framecount;
  double TimeCount = 0;
  double rem;
  int i, tmp;
  double maxOflow=0;
  double tmpOflow;
  int slotno = 0;
  double rate, time;
  char startDisplay = 0;
  double slotOverflow;
  double slotOverflowTotal;
  double framesBuffered;

  fseek(fd,0,0);

  Framecount = 0;
  Uflow *= -1;

  TimeCount = 0;
  slotOverflowTotal = 0;
  framesBuffered = 0;

  while (!feof(fd)) {                 
     if (TimeCount < sendInterval) {    
        fscanf(fd, "%d %d %lf %lf\n", &fno, &fsize, &rate, &time);
        Framecount += 1.0;
        TimeCount += time;

        Qptr = (struct Queue *)malloc(sizeof(struct Queue));
        Qptr->fsize = fsize;
        Qptr->fno = fno;
        Qptr->next = NULL;

        if (Qhead == NULL) {
           Qhead = Qptr;
           Current = Qptr;
        } else {
           Current->next = Qptr;
           Current = Qptr;
        } 
     }
     if (TimeCount >= sendInterval) {     /* one time slot has elapsed */
        rem = TimeCount - sendInterval;
        if (rem > 0) {
            Framecount -= 1.0;    /* Queue element is actually only partly */ 
            PartFrameptr = Qptr;  /*   delivered. Can't be consumed yet.   */
        } else {                  
            PartFrameptr = NULL;  /* Entire queue can be consumed, the */
        }                         /* entire frame has now been sent. */
        /* 
         * Check if ready to start display consumption. If so,
         * Measure Qlength and consume a slot.
         */
        if (startDisplay) {
           tmpOflow = getOverflow(maxOflow, rate, rem);

           if (tmpOflow > maxOflow) {
              maxOflow = tmpOflow; 
	   }
           if (consumeSlot(PartFrameptr) < 0) {
              printf("Warning!! Underflow Condition\n");
              return(-125);
           }
        } 
        /*
         * Check if Uflow has been filled. If so, start display
         * after appropriate slots have been buffered. We must
         * buffer the overflow requirement, PLUS the number of
         * frames in the buffered slots that contributed to the
         * overflow factor in underflow computation.  
         */
        if (!startDisplay) {
           framesBuffered += Framecount;
           slotOverflow = Framecount - framesPerInterval;
           if (slotOverflow > 0) {
              slotOverflowTotal += slotOverflow;
	    }

           if (framesBuffered >= Uflow) {  
              if ((Uflow + slotOverflowTotal) <= framesBuffered) {
                  startDisplay = 1; 
	      }
           } 
        }
        /*
         * Set-up for next time slot.
         * Account for previous slot spill over.
         */
        TimeCount = rem;
        if (rem > 0) Framecount = 1;
        else Framecount = 0;
     } 
   }  /* end while */ 

   while (Qhead != NULL) {  /* Empty up the queue. No more deliveries */
      consumeSlot(NULL);    /*   from the server.                     */
   }
   return(maxOflow); 
}

/*--------------------------*/
double getOverflow(oflow, rate, rem)
   double oflow, rate, rem;
{
   struct Queue *Qptr;
   double over = 0;

   Qptr = Qhead;
   if (Qhead == NULL) {
      return 0;
    } else {
       while (Qptr != NULL) {
          over += Qptr->fsize * 1.0;           
          Qptr = Qptr->next;
       }
       over -= rate * rem;  /* substract bits to be pushed over */
                            /* to the next time slot.           */
#ifdef DEBUG
  printf("###### over: %e,  pushed %e\n", over, rate*rem); 
#endif
       return over;
    }
}


/*--------------------------*/
int consumeSlot(PartFrameptr) 
   struct Queue *PartFrameptr;
{
   double i;
   struct Queue *Qptr;
 
   i = 0;
   while (i < framesPerInterval) {
      if ((Qhead == NULL) || (Qhead == PartFrameptr)) { 
         return -1;
      }
      Qptr = Qhead;
      Qhead = Qptr->next;
      free(Qptr);
      ConsumedFrames++;
      i += 1.0;
    }
} /* end consumeSlot() */ 



